/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: adi2send.c,v 1.1.1.1 1997/09/17 20:39:22 gropp Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpid.h"
#include "mpiddev.h"
#include "chgamma.h"
#ifdef USE_GAMMA_ISEND
#include "memlock.h"
#endif

/***************************************************************************/
/*
 * Multi-protocol, Multi-device support for 2nd generation ADI.
 * We start with support for blocking, contiguous sends.
 * Note the 'msgrep' field; this gives a hook for heterogeneous systems
 * which can be ignored on homogeneous systems.
 */
/***************************************************************************/


void MPID_SendContig (
	struct MPIR_COMMUNICATOR *comm_ptr, void *buf, int len,
	int src_lrank, int tag, int context_id, int dest_grank,
	MPID_Msgrep_t msgrep, int *error_code
) {
MPID_Device *dev;
int (*fcn) ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t ));

dev = MPID_devset->dev[dest_grank];

/* The one error test that makes sense here */

if (buf != 0 || len == 0) {

	/* Choose the function based on the message length in bytes */
	if (len < dev->long_len)
		fcn = dev->short_msg->send;
#if 0
	else if (len < dev->vlong_len && MPID_FLOW_MEM_OK(len,dest_grank)) 
		fcn = dev->long_msg->send;
#endif
	else
		fcn = dev->vlong_msg->send;

	DEBUG_TEST_FCN(fcn,"dev->proto->send");
	*error_code = (*(fcn))(
		buf, len, src_lrank, tag, context_id, dest_grank, msgrep
	);

	return;
}
*error_code = MPI_ERR_BUFFER;
return;
}


void MPID_IsendContig(
	struct MPIR_COMMUNICATOR *comm_ptr, void *buf, int len,
	int src_lrank, int tag, int context_id, int dest_grank,
	MPID_Msgrep_t msgrep, MPI_Request request, int *error_code
) {
	MPID_Device *dev = MPID_devset->dev[dest_grank];
	int (*fcn) ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t, 
		MPIR_SHANDLE * ));

	/* The one error test that makes sense here */
	if (buf == 0 && len > 0) {
		*error_code = MPI_ERR_BUFFER;
		return;
	}

	request->shandle.finish = 0;

	/* Choose the function based on the message length in bytes */
	if (dest_grank != MPID_MyWorldRank) {
		if (len < dev->long_len)
			fcn = dev->short_msg->isend;
#if 0
		else if (len < dev->vlong_len && MPID_FLOW_MEM_OK(len,dest_grank)) 
			fcn = dev->long_msg->isend;
#endif
		else
			fcn = dev->vlong_msg->isend;
	}
	else
		fcn = dev->rndv->isend;

#ifdef USE_GAMMA_ISEND
	AG_mlock(buf,len);
	/*  These are also for later AG_munlock()  */
#endif
	request->shandle.start = buf;
	request->shandle.bytes_as_contig = len;

	DEBUG_TEST_FCN(fcn,"dev->proto->isend");
	*error_code = (*(fcn))(
		buf, len, src_lrank, tag, context_id, dest_grank, msgrep,
		(MPIR_SHANDLE *)request
	);
	return;
}


#if 0
/* Bsend is just a test for short send */
void MPID_BsendContig(
	struct MPIR_COMMUNICATOR *comm_ptr, void *buf, int len,
	int src_lrank, int tag, int context_id, int dest_grank,
	MPID_Msgrep_t msgrep, int *error_code
) {
	MPID_Device *dev = MPID_devset->dev[dest_grank];
	int rc;

	if (len < dev->long_len) {
		AG_mlock(buf,len);
		DEBUG_TEST_FCN(dev->short_msg->send,"dev->short->send");
		rc = (*dev->short_msg->send)(
			buf, len, src_lrank, tag, context_id, dest_grank,
			msgrep
		);
		AG_munlock(buf,len);
	}
	else
		rc = MPIR_ERR_MAY_BLOCK;
	*error_code = rc;
}
#endif


int MPID_SendIcomplete(
	MPI_Request request, int *error_code
) {
	MPIR_SHANDLE *shandle = &request->shandle;
	MPID_Device *dev;
	int lerr;

	*error_code = MPI_SUCCESS;

	if (shandle->test) {
		lerr = (*shandle->test)( shandle );
		if (lerr < 0) {
			/*  Operation not yet complete  */
			return 0;
		}
		if (lerr == 0) {
			/*
			 *  Operation completed successfully.
			 *  shandle->is_complete has been set to 1 by the
			 *  test() function.
			 */
			if (shandle->finish) (shandle->finish)( shandle );
#ifdef USE_GAMMA_ISEND
			AG_munlock(shandle->start,shandle->bytes_as_contig);
#endif
			return 1;
		}
		/*  Operation failed, error --> status  */
		*error_code = lerr;
		return 0;
	}

	if (!shandle->is_complete) {
		dev = MPID_devset->dev_list;
		lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
		if (lerr < 0) {
			return 0;
		}
		if (lerr > 0) {
			*error_code = lerr;
			return 0;
		}
		if (!shandle->is_complete) {
			return 0;
		}
	}

	if (shandle->finish) (shandle->finish)( shandle );
#ifdef USE_GAMMA_ISEND
	AG_munlock(shandle->start,shandle->bytes_as_contig);
#endif
	return 1;
}


void MPID_SendComplete(
	MPI_Request request, int *error_code
) {
	MPIR_SHANDLE *shandle = &request->shandle;
	int          lerr;

	/*
	The 'while' is at the top in case the 'wait' routine is changed
	by one of the steps.  This happens, for example, in the Rendezvous
	Protocol
	*/
	while (!shandle->is_complete) {
		if (shandle->wait)
			*error_code = (*shandle->wait)( shandle );
		else {
			/*
			The most common case is a check device loop until it is
			complete.
			*/
			MPID_Device *dev;

			if (MPID_devset->ndev_list == 1) {
				dev = MPID_devset->dev_list;
				if (!shandle->is_complete) {
					lerr = (*dev->check_device)(
						dev, MPID_BLOCKING
					);
					if (lerr > 0) {
						*error_code = lerr;
						return;
					}
				}
			}
			else {
				if (!shandle->is_complete) {
					dev = MPID_devset->dev_list;
					while (dev) {
						lerr = (*dev->check_device)(
							dev, MPID_NOTBLOCKING
						);
						if (lerr > 0) {
							*error_code = lerr;
							return;
						}
						dev = dev->next;
					}
				}
			}
		}
	}
	if (shandle->finish) (shandle->finish)( shandle );
#ifdef USE_GAMMA_ISEND
	AG_munlock(shandle->start,shandle->bytes_as_contig);
#endif
	return;
}
