/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: chchkdev.c,v 1.1.1.1 1997/09/17 20:39:24 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpid.h"
#include "mpiddev.h"
#include "queue.h"
#include "reqalloc.h"
#include "chgamma.h"
#include "threadsafe.h"

/***************************************************************************/
/* This is one of the main routines.  It checks for incoming messages and  */
/* dispatches them.  There is another such look in MPID_CH_blocking_recv   */
/* which is optimized for the important case of blocking receives for a    */
/* particular message.                                                     */
/***************************************************************************/

/* Check for incoming messages.
    Input Parameter:
.   is_blocking - true if this routine should block until a message is
    available

    Returns -1 if nonblocking and no messages pending.  Otherwise 
    returns error code (MPI_SUCCESS == 0 for success)

    This routine makes use of a single dispatch routine to handle all
    incoming messages.  This makes the code a little lengthy, but each
    piece is relatively simple.

    This is the message-passing version.  The shared-memory version is
    in chchkshdev.c .
 */    
int MPID_CH_Check_incoming( MPID_Device *dev, MPID_BLOCKING_TYPE is_blocking )
{
	MPID_PKT_HEAD_T	*gpkt;
	unsigned int	tail;
	int		from_grank;
	MPIR_RHANDLE	*rhandle;
	int		unexp_entry;
	int		err = MPI_SUCCESS;

#if AG_DEBUG
	printf("[%d] chkdev(): Entering\n", gamma_my_node());
	fflush(stdout);
#endif

	/* If nonblocking and no headers available, exit */
	if (is_blocking == MPID_NOTBLOCKING) {
		if (gamma_test(AG_NEWMSG_SEM) == 0) {
			DEBUG_PRINT_MSG("Leaving check_incoming (no messages)");
			return -1;
		}
		DEBUG_PRINT_MSG("Message is available!");
	}
#if AG_DEBUG
	printf("[%d] chkdev(): Waiting for msg to arrive\n",gamma_my_node());
	fflush(stdout);
#endif

	if (gamma_wait(AG_NEWMSG_SEM,1) == 0) {

	/*  There is a new message.  Wait for its completion...  */

        /*  Thread-safe increment of tail index  */
	AG_spinlock_take();
	tail = AG_tail;
	AG_tail = (AG_tail + 1) % AG_MAXQLEN;
	AG_spinlock_release();

	from_grank = AG_src_queue[tail].src;
#if AG_DEBUG
	printf("[%d] chkdev(): Msg from node %d.  Wait for completion\n",
		gamma_my_node(), from_grank);
	fflush(stdout);
#endif

	if (gamma_wait(from_grank,1) == 0) {

	unexp_entry = AG_src_queue[tail].unexp_entry;
	gpkt = (MPID_PKT_HEAD_T *)&(AG_pkt_queue[tail].p);

#if AG_DEBUG
	printf("[%d] chkdev(): tail %ld, len %d, unexp_entry %d\n",
		gamma_my_node(), tail, gpkt->len, unexp_entry);
	fflush(stdout);
#endif

	/* 
	This unpacks ONLY the head of the message.
	Note that the payload is handled separately (MPIR_Unpack etc) and
	most of the other data can be considered just bits to return 
	uninterpreted. 
	There are exceptions (see rendevous code); 
	*/
	MPID_PKT_UNPACK( gpkt, sizeof(MPID_PKT_HEAD_T), from_grank );

	DEBUG_PRINT_PKT("R received message",&pkt);

	/* Is this message expected or not? */

	if (unexp_entry < 0) {
		/*  Message was expected.  */
		rhandle = AG_src_queue[tail].rhandle;
		MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)gpkt->msgrep);
		switch (gpkt->mode) {
			case MPID_PKT_SHORT:
#if AG_DEBUG
		printf("[%d] chkdev(): short expected\n",gamma_my_node());
		fflush(stdout);
#endif
		DEBUG_TEST_FCN(dev->short_msg->recv,"dev->short->recv");
		err = (*dev->short_msg->recv)( rhandle, from_grank, gpkt );
		break;

			case MPID_PKT_REQUEST_SEND:
#if AG_DEBUG
{
MPID_PKT_T *pkt = (MPID_PKT_T *)gpkt;
printf("[%d] expected REQ: tag %d, context %d send_id %ld\n",
gamma_my_node(),
pkt->request_pkt.tag,
pkt->request_pkt.context_id,
(unsigned long)AG_src_queue[tail].temp_buf
);
fflush(stdout);
}
#endif
		DEBUG_TEST_FCN(dev->rndv->irecv,"dev->rndv->irecv");
		AGR_send_id = (unsigned long)AG_src_queue[tail].temp_buf;
		err = (*dev->rndv->irecv)( rhandle, from_grank, gpkt );
		break;

			case MPID_PKT_OK_TO_SEND:
#if AG_DEBUG
{
MPID_PKT_T *pkt = (MPID_PKT_T *)gpkt;
printf("[%d] expected OK: send_id %d\n",
gamma_my_node(), (int)pkt->sendok_pkt.send_id);
fflush(stdout);
}
#endif
		DEBUG_TEST_FCN(dev->rndv->do_ack,"dev->rndv->do_ack");
		err = (*dev->rndv->do_ack)( gpkt, from_grank );
		break;
    
			case MPID_PKT_LONG:
		/*  This is the completion of a rndv communication.  */
#if AG_DEBUG
		printf("[%d] chkdev(): rndv expected\n",gamma_my_node());
		fflush(stdout);
#endif
		DEBUG_TEST_FCN(dev->short_msg->recv,"dev->short->recv");
		err = (*dev->short_msg->recv)( rhandle, from_grank, gpkt );
		break;
    
			default:
		fprintf( stderr, 
			 "[%d] Internal error: msg packet ddiscarded (%s:%d)\n",
			 MPID_MyWorldRank, __FILE__, __LINE__ );
		fflush( stderr );
		}
	}
	else {
		/*  Unexpected message.  */
		/*
		 *  The entry no. "unexp_entry" of unexpected queue
		 *  has already been prepared by the GAMMA receiver
		 *  handler, and contains
		 *  all the data relevant to match a receive operation
		 *  (except for OK_TO_SEND messages, relative to freed
		 *  send requests; in this case "unexp_entry" carries
		 *  no additional information).
		 *  Now the entry must be completed:
		 *  Locate the unexp. entry, allocate a handle, then
		 *  put a handle pointer into the unexp. entry.
		 */
		/*
		 *  Allocate and fill up a recv handle.
		 *  Note that we don't initialize the request completely:
		 *  we init just the basic information on the request;
		 *  all other fields will be set when the message is extracted.
		 */
		MPID_RecvAlloc( rhandle );
		if (!rhandle) {
			MPIR_ERROR(
			MPIR_COMM_WORLD, MPI_ERR_INTERN,
			"Could not dynamically allocate rhandle"
			);
		}
		rhandle->s.MPI_SOURCE = gpkt->lrank;
		rhandle->s.MPI_TAG    = gpkt->tag;
		/* Note that we don't save the context_id or set a datatype */
		rhandle->is_complete  = 0;
		/*
		 *  Put rhandle ptr into the relevant entry in unexp. queue.
		 */
		MPID_recvs.unexpected.queue[unexp_entry].ptr = rhandle;
		MPID_DO_HETERO(
		rhandle->msgrep = (MPID_Msgrep_t)gpkt->head.msgrep
		);
		switch (gpkt->mode) {
			case MPID_PKT_SHORT:
#if AG_DEBUG
printf("[%d] chkdev(): short unexpected. Temp buffered at %p, len %d\n",
gamma_my_node(),AG_src_queue[tail].temp_buf,gpkt->len);
MPID_Dump_queues();
fflush(stdout);
#endif
		DEBUG_TEST_FCN(dev->short_msg->unex,"dev->short->unex");
		AG_tempbuf_ptr = AG_src_queue[tail].temp_buf;
		err = (*dev->short_msg->unex)( rhandle, from_grank, gpkt );
		break;

			case MPID_PKT_REQUEST_SEND:
#if AG_DEBUG
{
MPID_PKT_T *pkt = (MPID_PKT_T *)gpkt;
printf("[%d] unexpected REQ: tag %d, context %d send_id %ld\n",
gamma_my_node(),
pkt->request_pkt.tag,
pkt->request_pkt.context_id,
(unsigned long)AG_src_queue[tail].temp_buf
);
MPID_Dump_queues();
fflush(stdout);
}
#endif
		DEBUG_TEST_FCN(dev->rndv->unex,"dev->rndv->unex");
		AGR_send_id = (unsigned long)AG_src_queue[tail].temp_buf;
		err = (*dev->rndv->unex)( rhandle, from_grank, gpkt );
		break;

			case MPID_PKT_OK_TO_SEND:
#if AG_DEBUG
{
MPID_PKT_T *pkt = (MPID_PKT_T *)gpkt;
printf("[%d] expected OK for a freed send request: send_id %d\n",
gamma_my_node(), (int)pkt->sendok_pkt.send_id);
fflush(stdout);
}
#endif
		/*
		 *  This is an OK_TO_SEND ctl message whose related send
		 *  request has been withdrawn meanwhile, an unfortunate
		 *  and hopefully unfrequent case.
		 *  The workaround should work, but doesn't look very elegant.
		 */
		/* peppe: qui e' rattoppata, cerca di migliorare  */
		MPID_RecvFree( rhandle );  /* peppe: si poteva evitare... */
		DEBUG_TEST_FCN(dev->rndv->do_ack,"dev->rndv->do_ack");
		err = MPID_CH_Rndvb_ack_freed( gpkt, from_grank );
		break;
	    
/*
		This should never occur in this implementation!
			case MPID_PKT_LONG:
		DEBUG_TEST_FCN(dev->eager->unex,"dev->eager->unex");
		err = (*dev->eager->unex)( rhandle, from_grank, gpkt );
		break;
*/
	    
			default:
		fprintf( stderr, 
			 "[%d] Internal error: msg packet discarded (%s:%d)\n",
			 MPID_MyWorldRank, __FILE__, __LINE__ );
		}
	}
#if AG_DEBUG
	printf("[%d] chkdev(): Exiting\n", gamma_my_node());
	fflush(stdout);
#endif
	return err;

	}
	}

/* peppe qui perfezionare err management */
#if AG_DEBUG
	printf("[%d] chkdev(): recv internal error %d %d\n",
		gamma_my_node(), gamma_test(AG_1), gamma_test(AG_2));
	MPID_Dump_queues();
	fflush(stdout);
#endif
	MPID_Abort(
		(struct MPIR_COMMUNICATOR *)0, 1, "MPI internal",
		"Fatal recv internal err");
	return MPI_ERR_INTERN;
}
